use hifitime::{leap_seconds::{LeapSecond, LeapSecondProvider}, Epoch, Unit, TimeScale};

#[derive(Clone)]
struct LeapTable {
    table: Vec<LeapSecond>,
    pos: usize,
}

impl LeapTable {
    fn new(table: Vec<LeapSecond>) -> Self {
        Self { table, pos: 0 }
    }
}

impl Iterator for LeapTable {
    type Item = LeapSecond;

    fn next(&mut self) -> Option<Self::Item> {
        self.pos += 1;
        self.table.get(self.pos - 1).copied()
    }
}

impl DoubleEndedIterator for LeapTable {
    fn next_back(&mut self) -> Option<Self::Item> {
        if self.pos == self.table.len() {
            None
        } else {
            self.pos += 1;
            self.table.get(self.table.len() - self.pos).copied()
        }
    }
}

impl std::ops::Index<usize> for LeapTable {
    type Output = LeapSecond;

    fn index(&self, index: usize) -> &Self::Output {
        self.table.index(index)
    }
}

impl LeapSecondProvider for LeapTable {}

// This matches the behavior of the Epoch::from_utc_seconds function, but lets
// us use our own leap second provider.
fn from_utc_seconds<L>(seconds: f64, leaps: L) -> Epoch
    where L: LeapSecondProvider
{
    let mut e = Epoch::from_tai_duration(seconds * Unit::Second);
    e.duration_since_j1900_tai += e.leap_seconds_with(true, leaps).unwrap_or(0.0) * Unit::Second;
    e.time_scale = TimeScale::UTC;
    e
}

// This matches the behavior of the Epoch::to_utc_seconds function, but lets
// us use our own leap second provider.
fn to_utc_seconds<L>(e: &Epoch, leaps: L) -> f64
    where L: LeapSecondProvider
{
    let d = e.duration_since_j1900_tai - e.leap_seconds_with(true, leaps).unwrap_or(0.0) * Unit::Second;
    d.to_unit(Unit::Second)
}

fn fixed_to_utc<L>(e: &Epoch, leaps: L) -> f64
    where L: LeapSecondProvider
{
    let l = e.duration_since_j1900_tai.to_seconds();
    for leap_second in leaps.rev() {
        let maybe_converted = l - leap_second.delta_at;
        if maybe_converted >= leap_second.timestamp_tai_s {
            return maybe_converted;
        }
    }
    l
}

fn main() {

    let leap_insert = LeapTable::new(vec![
        LeapSecond::new(50.0, 3.0, true),
        LeapSecond::new(100.0, 4.0, true),
    ]);
    let leap_delete = LeapTable::new(vec![
        LeapSecond::new(50.0, 3.0, true),
        LeapSecond::new(100.0, 2.0, true),
    ]);

    println!();
    println!("Insert leap second @ 100, from +3 to +4");
    println!("UTC -> TAI -> UTC");
    for i in 96..=102 {
        let time = from_utc_seconds(i as f64, leap_insert.clone());
        let tai = time.to_tai_seconds().round() as i64;
        let utc = to_utc_seconds(&time, leap_insert.clone()).round() as i64;
        println!("{:3} -> {:3} -> {:3}, delta = {}", i, tai, utc, utc-i);
        if i == 99 {
            let time = Epoch::from_tai_seconds(103.0);
            let utc = to_utc_seconds(&time, leap_insert.clone()).round() as i64;
            println!("    -> {:3} -> {:3}, delta = {}", 103, utc, 0);
        }
    }

    println!();
    println!("Delete leap second @ 100, from +3 to +2");
    println!("UTC -> TAI -> UTC");
    for i in 96..=102 {
        let time = from_utc_seconds(i as f64, leap_delete.clone());
        let tai = time.to_tai_seconds().round() as i64;
        let utc = to_utc_seconds(&time, leap_delete.clone()).round() as i64;
        println!("{:3} -> {:3} -> {:3}, delta = {}", i, tai, utc, utc-i);
    }

    println!();
    println!("=== Proposed fix ===");

    println!();
    println!("Insert leap second @ 100, from +3 to +4");
    println!("UTC -> TAI -> UTC");
    for i in 96..=102 {
        let time = from_utc_seconds(i as f64, leap_insert.clone());
        let tai = time.to_tai_seconds().round() as i64;
        let utc = fixed_to_utc(&time, leap_insert.clone()).round() as i64;
        println!("{:3} -> {:3} -> {:3}, delta = {}", i, tai, utc, utc-i);
        if i == 99 {
            let time = Epoch::from_tai_seconds(103.0);
            let utc = to_utc_seconds(&time, leap_insert.clone()).round() as i64;
            println!("    -> {:3} -> {:3}, delta = {}", 103, utc, 0);
        }
    }

    println!();
    println!("Delete leap second @ 100, from +3 to +2");
    println!("UTC -> TAI -> UTC");
    for i in 96..=102 {
        let time = from_utc_seconds(i as f64, leap_delete.clone());
        let tai = time.to_tai_seconds().round() as i64;
        let utc = fixed_to_utc(&time, leap_delete.clone()).round() as i64;
        println!("{:3} -> {:3} -> {:3}, delta = {}", i, tai, utc, utc-i);
    }
}
